"""
Currex - A Python library for currency conversion and arithmetic
"""

from .currency import Currency
from .factory import (
    USD,
    EUR,
    GBP,
    JPY,
    CHF,
    AUD,
    CAD,
    CNY,
    HKD,
    NZD,
    SEK,
    KRW,
    SGD,
    NOK,
    MXN,
    INR,
    RUB,
    ZAR,
    TRY,
    BRL,
    TWD,
    DKK,
    PLN,
    THB,
    IDR,
    CZK,
    AED,
    AFN,
    AOA,
    ARS,
    BAM,
    BGN,
    BYN,
    CDF,
    CLP,
    COP,
    CUP,
    GEL,
    GHS,
    GNF,
    HUF,
    ILS,
    ISK,
    KZT,
    MDL,
    MGA,
    MRU,
    MZN,
    NIO,
    PEN,
    PHP,
    PKR,
    RON,
    RSD,
    SDG,
    SRD,
    STN,
    TJS,
    TMT,
    UAH,
    UGX,
    UYU,
    UZS,
    VES,
    XAF,
    XOF,
    ZMW,
)
from .exchange import ExchangeRateAPI

__version__ = "0.1.1"
__all__ = [
    "Currency",
    "ExchangeRateAPI",
    "USD",
    "EUR",
    "GBP",
    "JPY",
    "CHF",
    "AUD",
    "CAD",
    "CNY",
    "HKD",
    "NZD",
    "SEK",
    "KRW",
    "SGD",
    "NOK",
    "MXN",
    "INR",
    "RUB",
    "ZAR",
    "TRY",
    "BRL",
    "TWD",
    "DKK",
    "PLN",
    "THB",
    "IDR",
    "CZK",
    "AED",
    "AFN",
    "AOA",
    "ARS",
    "BAM",
    "BGN",
    "BYN",
    "CDF",
    "CLP",
    "COP",
    "CUP",
    "GEL",
    "GHS",
    "GNF",
    "HUF",
    "ILS",
    "ISK",
    "KZT",
    "MDL",
    "MGA",
    "MRU",
    "MZN",
    "NIO",
    "PEN",
    "PHP",
    "PKR",
    "RON",
    "RSD",
    "SDG",
    "SRD",
    "STN",
    "TJS",
    "TMT",
    "UAH",
    "UGX",
    "UYU",
    "UZS",
    "VES",
    "XAF",
    "XOF",
    "ZMW",
]



"""Tests for currency chain conversions.

This test file verifies that we can convert between pairs of currencies
by converting 1 unit of currency at index 2n to currency at index 2n+1.
"""

import pytest
from currex import (
    USD,
    EUR,
    GBP,
    JPY,
    CHF,
    AUD,
    CAD,
    HKD,
    NZD,
    SEK,
    KRW,
    SGD,
    NOK,
    MXN,
    INR,
    RUB,
    ZAR,
    TRY,
    BRL,
    TWD,
    DKK,
    PLN,
    THB,
    IDR,
    CZK,
    AED,
    AFN,
    AOA,
    ARS,
    BAM,
    BGN,
    BYN,
    CDF,
    CLP,
    COP,
    CUP,
    GEL,
    GHS,
    GNF,
    HUF,
    ILS,
    ISK,
    KZT,
    MDL,
    MGA,
    MRU,
    MZN,
    NIO,
    PEN,
    PHP,
    PKR,
    RON,
    RSD,
    SDG,
    SRD,
    STN,
    TJS,
    TMT,
    UAH,
    UGX,
    UYU,
    UZS,
    VES,
    XAF,
    XOF,
    ZMW,
)


@pytest.mark.api
def test_specific_currency_pairs():
    """Test some specific currency pairs that are commonly used."""
    test_cases = [
        (USD(1), EUR),  # USD to EUR
        (EUR(1), GBP),  # EUR to GBP
        (GBP(1), JPY),  # GBP to JPY
        (CHF(1), CAD),  # CHF to CAD
        (AUD(1), SGD),  # AUD to SGD
    ]

    for amount, target_currency in test_cases:
        converted = amount.to(target_currency)
        assert isinstance(converted, target_currency)
        assert (
            converted.amount > 0
        ), f"Converting 1 {amount.__class__.__name__} to {target_currency.__name__} resulted in 0 or negative amount"
        print(f"1 {amount.__class__.__name__} = {converted}")


@pytest.mark.api
def test_currency_conversion_all():
    """Test converting 1 unit between pairs of currencies."""
    # Create a list of all currency classes
    currencies = [
        USD,
        EUR,
        GBP,
        JPY,
        CHF,
        AUD,
        CAD,
        HKD,
        NZD,
        SEK,
        KRW,
        SGD,
        NOK,
        MXN,
        INR,
        RUB,
        ZAR,
        TRY,
        BRL,
        TWD,
        DKK,
        PLN,
        THB,
        IDR,
        CZK,
        AED,
        AFN,
        AOA,
        ARS,
        BAM,
        BGN,
        BYN,
        CDF,
        CLP,
        COP,
        CUP,
        GEL,
        GHS,
        GNF,
        HUF,
        ILS,
        ISK,
        KZT,
        MDL,
        MGA,
        MRU,
        MZN,
        NIO,
        PEN,
        PHP,
        PKR,
        RON,
        RSD,
        SDG,
        SRD,
        STN,
        TJS,
        TMT,
        UAH,
        UGX,
        UYU,
        UZS,
        VES,
        XAF,
        XOF,
        ZMW,
    ]

    # Convert 1 unit from currency at index 2n to currency at index 2n+1
    for i in range(0, len(currencies) - 1, 2):
        from_currency = currencies[i]
        to_currency = currencies[i + 1]

        # Create amount of 1 in the source currency
        amount = from_currency(1)

        # Convert to target currency
        converted = amount.to(to_currency)

        # Verify the conversion produced a valid result
        assert isinstance(converted, to_currency)
        assert (
            converted.amount > 0
        ), f"Converting 1 {from_currency.__name__} to {to_currency.__name__} resulted in 0 or negative amount"

        print(f"1 {from_currency.__name__} = {converted}")



from decimal import Decimal
from typing import Union, Type, TypeVar
from .exchange import ExchangeRateAPI

DecimalLike = Union[int, float, Decimal]
C = TypeVar("C", bound="Currency")


def to_decimal(value: DecimalLike) -> Decimal:
    """Convert a DecimalLike value to Decimal"""
    return value if isinstance(value, Decimal) else Decimal(str(value))


class CurrencyMeta(type):
    def __mul__(cls: Type[C], other: DecimalLike) -> C:
        return cls(to_decimal(other))

    def __rmul__(cls: Type[C], other: DecimalLike) -> C:
        return cls(to_decimal(other))


class Currency(metaclass=CurrencyMeta):
    def __init__(self, amount: Union["Currency", DecimalLike]):
        if isinstance(amount, Currency):
            converted = amount.to(type(self))
            self.amount = converted.amount
        else:
            self.amount = to_decimal(amount)

    def __mul__(self: C, other: DecimalLike) -> C:
        return type(self)(self.amount * to_decimal(other))

    def __rmul__(self: C, other: DecimalLike) -> C:
        return self.__mul__(other)

    def __add__(self: C, other: Union["Currency", DecimalLike]) -> C:
        if isinstance(other, Currency):
            converted = other.to(type(self))
            return type(self)(self.amount + converted.amount)
        return type(self)(self.amount + to_decimal(other))

    def __radd__(self: C, other: DecimalLike) -> C:
        return type(self)(self.amount + to_decimal(other))

    def __neg__(self: C) -> C:
        """Return the negative of this currency amount"""
        return type(self)(-self.amount)

    def __sub__(self: C, other: Union["Currency", DecimalLike]) -> C:
        """Subtract another currency (with conversion) or decimal-like number"""
        if isinstance(other, Currency):
            converted = other.to(type(self))
            return type(self)(self.amount - converted.amount)
        return type(self)(self.amount - to_decimal(other))

    def __rsub__(self: C, other: DecimalLike) -> C:
        """Subtract this currency from a decimal-like number"""
        return type(self)(to_decimal(other) - self.amount)

    def __truediv__(self, other: Union["Currency", DecimalLike]) -> Union[C, float]:
        """Divide by another currency (returns unitless) or decimal-like number (returns same currency)"""
        if isinstance(other, Currency):
            converted = other.to(type(self))
            return float(self.amount / converted.amount)
        return type(self)(self.amount / to_decimal(other))

    def __eq__(self, other: object) -> bool:
        """Equal only if same currency type and amount"""
        if not isinstance(other, Currency):
            return NotImplemented
        return type(self) is type(other) and self.amount == other.amount

    def __ne__(self, other: object) -> bool:
        """Not equal if different currency type or amount"""
        if not isinstance(other, Currency):
            return NotImplemented
        return not self.__eq__(other)

    def __lt__(self, other: "Currency") -> bool:
        """Less than, converts other currency if needed"""
        if not isinstance(other, Currency):
            return NotImplemented
        if type(self) is not type(other):
            other = other.to(type(self))
        return self.amount < other.amount

    def __le__(self, other: "Currency") -> bool:
        """Less than or equal, converts other currency if needed"""
        if not isinstance(other, Currency):
            return NotImplemented
        if type(self) is not type(other):
            other = other.to(type(self))
        return self.amount <= other.amount

    def __gt__(self, other: "Currency") -> bool:
        """Greater than, converts other currency if needed"""
        if not isinstance(other, Currency):
            return NotImplemented
        if type(self) is not type(other):
            other = other.to(type(self))
        return self.amount > other.amount

    def __ge__(self, other: "Currency") -> bool:
        """Greater than or equal, converts other currency if needed"""
        if not isinstance(other, Currency):
            return NotImplemented
        if type(self) is not type(other):
            other = other.to(type(self))
        return self.amount >= other.amount

    def __str__(self) -> str:
        return f"{self.amount:.2f} {self.__class__.__name__}"

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.amount})"

    def to(self, currency_class: Type[C]) -> C:
        """Convert to another currency"""
        from_currency = self.__class__.__name__
        to_currency = currency_class.__name__
        if from_currency == to_currency:
            return currency_class(self.amount)

        rate = ExchangeRateAPI.get_rate(from_currency, to_currency)
        converted_amount = self.amount * rate
        return currency_class(converted_amount)



from decimal import Decimal
import requests


class ExchangeRateAPI:
    _api_url: str = "https://hexarate.paikama.co/api/rates/latest"
    _rates_cache: dict[str, Decimal] = {}

    @classmethod
    def get_rate(cls, from_currency: str, to_currency: str) -> Decimal:
        """Get exchange rate from Hexarate API"""
        cache_key = f"{from_currency}-{to_currency}"
        if cache_key in cls._rates_cache:
            return cls._rates_cache[cache_key]

        try:
            response = requests.get(
                f"{cls._api_url}/{from_currency}", params={"target": to_currency}
            )
            response.raise_for_status()
            data = response.json()
            if data["status_code"] != 200:
                raise ValueError(f"API returned error status: {data['status_code']}")
            rate = Decimal(str(data["data"]["mid"]))
            cls._rates_cache[cache_key] = rate
            return rate
        except Exception as e:
            raise ValueError(f"Failed to get exchange rate: {str(e)}")



from typing import Type
from .currency import Currency


def _currency_factory(code: str) -> Type[Currency]:
    """Creates a new currency class for the given currency code."""

    class NewCurrency(Currency):
        pass

    NewCurrency.__name__ = code
    return NewCurrency


# Create currency classes for major currencies
USD = _currency_factory("USD")  # United States Dollar
EUR = _currency_factory("EUR")  # Euro
GBP = _currency_factory("GBP")  # British Pound Sterling
JPY = _currency_factory("JPY")  # Japanese Yen
CHF = _currency_factory("CHF")  # Swiss Franc
AUD = _currency_factory("AUD")  # Australian Dollar
CAD = _currency_factory("CAD")  # Canadian Dollar
CNY = _currency_factory("CNY")  # Chinese Yuan
HKD = _currency_factory("HKD")  # Hong Kong Dollar
NZD = _currency_factory("NZD")  # New Zealand Dollar
SEK = _currency_factory("SEK")  # Swedish Krona
KRW = _currency_factory("KRW")  # South Korean Won
SGD = _currency_factory("SGD")  # Singapore Dollar
NOK = _currency_factory("NOK")  # Norwegian Krone
MXN = _currency_factory("MXN")  # Mexican Peso
INR = _currency_factory("INR")  # Indian Rupee
RUB = _currency_factory("RUB")  # Russian Ruble
ZAR = _currency_factory("ZAR")  # South African Rand
TRY = _currency_factory("TRY")  # Turkish Lira
BRL = _currency_factory("BRL")  # Brazilian Real
TWD = _currency_factory("TWD")  # New Taiwan Dollar
DKK = _currency_factory("DKK")  # Danish Krone
PLN = _currency_factory("PLN")  # Polish Złoty
THB = _currency_factory("THB")  # Thai Baht
IDR = _currency_factory("IDR")  # Indonesian Rupiah
CZK = _currency_factory("CZK")  # Czech Koruna
AED = _currency_factory("AED")  # UAE Dirham
AFN = _currency_factory("AFN")  # Afghan Afghani
AOA = _currency_factory("AOA")  # Angolan Kwanza
ARS = _currency_factory("ARS")  # Argentine Peso
BAM = _currency_factory("BAM")  # Bosnia-Herzegovina Convertible Mark
BGN = _currency_factory("BGN")  # Bulgarian Lev
BYN = _currency_factory("BYN")  # Belarusian Ruble
CDF = _currency_factory("CDF")  # Congolese Franc
CLP = _currency_factory("CLP")  # Chilean Peso
COP = _currency_factory("COP")  # Colombian Peso
CUP = _currency_factory("CUP")  # Cuban Peso
GEL = _currency_factory("GEL")  # Georgian Lari
GHS = _currency_factory("GHS")  # Ghanaian Cedi
GNF = _currency_factory("GNF")  # Guinean Franc
HUF = _currency_factory("HUF")  # Hungarian Forint
ILS = _currency_factory("ILS")  # Israeli New Shekel
ISK = _currency_factory("ISK")  # Icelandic Króna
KZT = _currency_factory("KZT")  # Kazakhstani Tenge
MDL = _currency_factory("MDL")  # Moldovan Leu
MGA = _currency_factory("MGA")  # Malagasy Ariary
MRU = _currency_factory("MRU")  # Mauritanian Ouguiya
MZN = _currency_factory("MZN")  # Mozambican Metical
NIO = _currency_factory("NIO")  # Nicaraguan Córdoba
PEN = _currency_factory("PEN")  # Peruvian Sol
PHP = _currency_factory("PHP")  # Philippine Peso
PKR = _currency_factory("PKR")  # Pakistani Rupee
RON = _currency_factory("RON")  # Romanian Leu
RSD = _currency_factory("RSD")  # Serbian Dinar
SDG = _currency_factory("SDG")  # Sudanese Pound
SRD = _currency_factory("SRD")  # Surinamese Dollar
STN = _currency_factory("STN")  # São Tomé and Príncipe Dobra
TJS = _currency_factory("TJS")  # Tajikistani Somoni
TMT = _currency_factory("TMT")  # Turkmenistani Manat
UAH = _currency_factory("UAH")  # Ukrainian Hryvnia
UGX = _currency_factory("UGX")  # Ugandan Shilling
UYU = _currency_factory("UYU")  # Uruguayan Peso
UZS = _currency_factory("UZS")  # Uzbekistani Som
VES = _currency_factory("VES")  # Venezuelan Bolívar Soberano
XAF = _currency_factory("XAF")  # CFA Franc BEAC
XOF = _currency_factory("XOF")  # CFA Franc BCEAO
ZMW = _currency_factory("ZMW")  # Zambian Kwacha

# Update __all__ to include all currency classes
__all__ = [
    "USD",
    "EUR",
    "GBP",
    "JPY",
    "CHF",
    "AUD",
    "CAD",
    "CNY",
    "HKD",
    "NZD",
    "SEK",
    "KRW",
    "SGD",
    "NOK",
    "MXN",
    "INR",
    "RUB",
    "ZAR",
    "TRY",
    "BRL",
    "TWD",
    "DKK",
    "PLN",
    "THB",
    "IDR",
    "CZK",
    "AED",
    "AFN",
    "AOA",
    "ARS",
    "BAM",
    "BGN",
    "BYN",
    "CDF",
    "CLP",
    "COP",
    "CUP",
    "GEL",
    "GHS",
    "GNF",
    "HUF",
    "ILS",
    "ISK",
    "KZT",
    "MDL",
    "MGA",
    "MRU",
    "MZN",
    "NIO",
    "PEN",
    "PHP",
    "PKR",
    "RON",
    "RSD",
    "SDG",
    "SRD",
    "STN",
    "TJS",
    "TMT",
    "UAH",
    "UGX",
    "UYU",
    "UZS",
    "VES",
    "XAF",
    "XOF",
    "ZMW",
]



import pytest
from currex import USD, EUR, GBP, PLN
from decimal import Decimal


def test_basic_multiplication():
    amount = 100 * USD
    assert isinstance(amount, USD)
    assert amount.amount == Decimal("100")

    amount = USD * 100
    assert isinstance(amount, USD)
    assert amount.amount == Decimal("100")


def test_basic_division():
    amount = USD(100) / 2
    assert isinstance(amount, USD)
    assert amount.amount == Decimal("50")


def test_addition():
    amount1 = 100 * USD
    amount2 = 50 * USD
    total = amount1 + amount2
    assert isinstance(total, USD)
    assert total.amount == Decimal("150")


def test_subtraction():
    amount1 = 100 * USD
    amount2 = 30 * USD
    diff = amount1 - amount2
    assert isinstance(diff, USD)
    assert diff.amount == Decimal("70")


def test_comparison():
    # Same currency comparisons
    amount1 = 100 * USD
    amount2 = 50 * USD
    amount3 = 100 * USD

    assert amount1 > amount2
    assert amount2 < amount1
    assert amount1 >= amount3
    assert amount1 <= amount3
    assert amount1 == amount3
    assert amount1 != amount2


def test_string_representation():
    amount = 42.42 * USD
    assert str(amount).endswith("USD")
    assert "42.42" in str(amount)


def test_negative_amounts():
    amount = -50 * USD
    assert amount.amount == Decimal("-50")

    amount2 = 30 * USD
    result = amount + amount2
    assert result.amount == Decimal("-20")


def test_zero_amount():
    amount = 0 * USD
    assert amount.amount == Decimal("0")



"""Tests for currency exchange functionality.

Note: Tests marked with @pytest.mark.api are skipped in CI environments (like GitHub Actions)
because they require access to external exchange rate API which:
1. May have rate limits or require authentication
2. May be unstable or slow, making CI less reliable
3. Would make tests dependent on external services

Run these tests locally with `pytest -v` or explicitly with `pytest -v -m api`
"""

import pytest
from currex import USD, EUR, GBP, PLN
from decimal import Decimal


@pytest.mark.api
def test_usd_to_eur_conversion():
    """Test USD to EUR conversion"""
    usd_amount = 100 * USD
    eur_amount = EUR(usd_amount)
    assert isinstance(eur_amount, EUR)
    assert eur_amount.amount > 0


@pytest.mark.api
def test_eur_to_pln_conversion():
    """Test EUR to PLN conversion"""
    eur_amount = 50 * EUR
    pln_amount = PLN(eur_amount)
    assert isinstance(pln_amount, PLN)
    assert pln_amount.amount > 0


@pytest.mark.api
def test_round_trip_conversion():
    """Test converting back and forth between currencies"""
    original = 100 * USD
    converted = EUR(original)
    back_to_usd = USD(converted)

    # Due to exchange rate fluctuations and rounding,
    # we check if the value is within 1% of original
    assert abs(back_to_usd.amount - original.amount) / original.amount < Decimal("0.01")


@pytest.mark.api
def test_multi_currency_chain():
    """Test converting through multiple currencies"""
    start_amount = 100 * USD
    # USD -> EUR -> GBP -> PLN
    result = PLN(GBP(EUR(start_amount)))
    assert isinstance(result, PLN)
    assert result.amount > 0


@pytest.mark.api
def test_comparison_different_currencies():
    """Test comparisons between different currencies (requires exchange rates)"""
    amount_pln = 100 * PLN
    amount_eur = 100 * EUR

    # Assuming 1 EUR is always worth more than 1 PLN
    assert amount_pln < amount_eur
    assert amount_pln <= amount_eur
